/*
 * Copyright (c) 2011 Michael Schierl
 * 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 
 * - Redistributions of source code must retain the above copyright notice,
 *   this list of conditions and the following disclaimer.
 *   
 * - Redistributions in binary form must reproduce the above copyright
 *   notice, this list of conditions and the following disclaimer in the
 *   documentation and/or other materials provided with the distribution.
 *   
 * - Neither name of the copyright holders nor the names of its
 *   contributors may be used to endorse or promote products derived from
 *   this software without specific prior written permission.
 *   
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
package net.sf.jmatchparser.util.charset.icu4jchardet;

import java.io.ByteArrayOutputStream;
import java.nio.ByteBuffer;
import java.nio.CharBuffer;
import java.nio.charset.Charset;
import java.nio.charset.CharsetDecoder;
import java.nio.charset.CharsetEncoder;
import java.nio.charset.CoderResult;
import java.nio.charset.UnsupportedCharsetException;

import com.ibm.icu.text.CharsetDetector;
import com.ibm.icu.text.CharsetMatch;

class ICU4JChardetCharset extends Charset {

	public static final String NAME = "ICU4J-Chardet";
	public static final String FILTERED_NAME = "ICU4J-Chardet-Filtered";

	private static ICU4JChardetCharset instance = null;
	private static ICU4JChardetCharset filteredInstance = null;

	protected static ICU4JChardetCharset getInstance() {
		if (instance == null) {
			instance = new ICU4JChardetCharset(false);
		}
		return instance;
	}
	
	protected static ICU4JChardetCharset getFilteredInstance() {
		if (filteredInstance == null) {
			filteredInstance = new ICU4JChardetCharset(true);
		}
		return filteredInstance;
	}
	
	private final boolean filtered;

	private ICU4JChardetCharset(boolean filtered) {
		super(NAME, null);
		this.filtered = filtered;
	}

	@Override
	public boolean contains(Charset cs) {
		return cs == this;
	}

	@Override
	public CharsetDecoder newDecoder() {
		return new Decoder();
	}

	@Override
	public CharsetEncoder newEncoder() {
		throw new UnsupportedOperationException();
	}

	@Override
	public boolean canEncode() {
		return false;
	}

	private class Decoder extends CharsetDecoder {
		private ByteArrayOutputStream buffer = new ByteArrayOutputStream();
		private CharsetDecoder usedDecoder = null;
		private ByteBuffer remaining = null;
		private boolean isFlushed = false;

		protected Decoder() {
			super(ICU4JChardetCharset.this, 1.0f, 2.0f);
		}


		@Override
		protected CoderResult decodeLoop(ByteBuffer in, CharBuffer out) {
			byte[] buf = new byte[1024];
			while (true) {
				if (in.remaining() == 0)
					return CoderResult.UNDERFLOW;
				int len = Math.min(in.remaining(), buf.length);
				in.get(buf, 0, len);
				buffer.write(buf, 0, len);
			}
		}

		@Override
		protected CoderResult implFlush(CharBuffer out) {
			if (usedDecoder == null) {
				CharsetDetector detector = new CharsetDetector();
				detector.enableInputFilter(filtered);
				byte[] data = buffer.toByteArray();
				detector.setText(data);
				CharsetMatch cm = detector.detect();
				try {
					usedDecoder = Charset.forName(cm == null ? "ISO-8859-1" : cm.getName()).newDecoder();
				} catch (UnsupportedCharsetException ex) {
					usedDecoder = Charset.forName("ISO-8859-1").newDecoder();
				}
				usedDecoder.onUnmappableCharacter(unmappableCharacterAction());
				usedDecoder.onMalformedInput(malformedInputAction());
				if (data.length > 0) {
					remaining = ByteBuffer.allocate(data.length);
					remaining.put(data);
					remaining.flip();
				}
				buffer = null;
			}
			if (remaining != null) {
				CoderResult result = usedDecoder.decode(remaining, out, true);
				if (remaining.remaining() == 0)
					remaining = null;
				if (!result.isUnderflow())
					return result;
			}
			if (!isFlushed) {
				CoderResult result = usedDecoder.flush(out);
				if (!result.isUnderflow())
					return result;
				isFlushed = true;
			}
			return super.implFlush(out);
		}

		@Override
		protected void implReset() {
			buffer = new ByteArrayOutputStream();
			usedDecoder = null;
			remaining = null;
			isFlushed = false;
		}

		@Override
		public boolean isAutoDetecting() {
			return true;
		}

		@Override
		public boolean isCharsetDetected() {
			return usedDecoder != null;
		}

		@Override
		public Charset detectedCharset() {
			return usedDecoder.charset();
		}
	}
}